-- Basic script to add reel/cutter effects to PhiBer Triple (mower)
-- Put together by The Man With No Name
-- July 2, 2018
-- Version 1.0.0.0

reelEffects = {}

function reelEffects.prerequisitesPresent(specializations)
	return SpecializationUtil.hasSpecialization(Mower, specializations)
end

function reelEffects:preLoad(savegame)
	self.loadTestAreaFromXML = reelEffects.loadTestAreaFromXML
    self.getIsTestAreaActive = reelEffects.getIsTestAreaActive
end

function reelEffects:load(savegame)
	if self.isClient then
		self.cutters = {}
		local j = 0
		while true do
			local key = string.format("vehicle.cutters.cutter(%d)", j)
			if not hasXMLProperty(self.xmlFile, key) then
				break
			end
			local entry = {}
			if hasXMLProperty(self.xmlFile, key..".turnedOnScrollers.turnedOnScroller") then
				entry.cutterTurnedOnScrollers = Utils.loadScrollers(self.components, self.xmlFile, key..".turnedOnScrollers.turnedOnScroller", {}, false)
			end
			if hasXMLProperty(self.xmlFile, key..".knifeAnimation") then
				entry.knifeAnimation = getXMLString(self.xmlFile, key..".knifeAnimation#animName")
				entry.knifeAnimationSpeed = Utils.getNoNil(getXMLFloat(self.xmlFile, key..".knifeAnimation#speed"), 1)
			end
			entry.threshingParticleSystems = {}
			local i = 0
			while true do
				local keyPS = string.format(key..".threshingParticleSystems.threshingParticleSystem(%d)", i)
				if not hasXMLProperty(self.xmlFile, keyPS) then
					break
				end
				local currentPS = {}
				ParticleUtil.loadParticleSystem(self.xmlFile, currentPS, keyPS, self.components, false, nil, self.baseDirectory)
				table.insert(entry.threshingParticleSystems, currentPS)
				i = i + 1
			end
			local i = 0
			while true do
				local keyES = string.format(key..".threshingParticleSystems.emitterShape(%d)", i)
				if not hasXMLProperty(self.xmlFile, keyES) then
					break
				end
				local emitterShape = Utils.indexToObject(self.components, getXMLString(self.xmlFile, keyES.."#node"))
				local particleType = getXMLString(self.xmlFile, key.."#particleType")
				if emitterShape ~= nil then
					local fillType = FillUtil.FILLTYPE_UNKNOWN
					local fillTypeStr = getXMLString(self.xmlFile, key.."#fillType")
					if fillTypeStr ~= nil and FillUtil.fillTypeNameToInt[fillTypeStr] ~= nil then
						fillType = FillUtil.fillTypeNameToInt[fillTypeStr]
					end
					local particleSystem = MaterialUtil.getParticleSystem(fillType, particleType)
					if particleSystem ~= nil then
						table.insert(entry.threshingParticleSystems, ParticleUtil.copyParticleSystem(self.xmlFile, key, particleSystem, emitterShape))
					end
				end
				i = i + 1
			end
			entry.cutterEffectBaseNode = Utils.getNoNil(Utils.indexToObject(self.components, getXMLString(self.xmlFile, key..".effect#baseNode")), self.rootNode)
			entry.cutterEffects = {}
			entry.currentCutterEffect = nil
			if hasXMLProperty(self.xmlFile, key..".effect") then
				for fruitTypeIndex,_ in pairs(FruitUtil.fruitTypeToWindrowFillType) do
					local centerEffect = MaterialUtil.getCutterEffectType("CENTER")
					local effectTable = MaterialUtil.getCutterEffects(fruitTypeIndex, centerEffect, true)
					if effectTable then
						entry.cutterEffects[fruitTypeIndex] = {}
						local i = 0
						while true do
							local keyEN = string.format(key..".effect.effectNode(%d)", i)
							if not hasXMLProperty(self.xmlFile, keyEN) then
								break
							end
							local effect = CutterEffect:new()
							effect:load(self.xmlFile, keyEN, self.components, self, fruitTypeIndex)
							table.insert(entry.cutterEffects[fruitTypeIndex], effect)
							i = i + 1
						end
					end
				end
			end
			entry.cutterTestAreas = {}
			local i = 0
			while true do
				local keyTA = string.format(key..".testAreas.testArea(%d)", i)
				if not hasXMLProperty(self.xmlFile, keyTA) then
					break
				end
				local testArea = {}
				if self:loadTestAreaFromXML(testArea, self.xmlFile, keyTA) then
					table.insert(entry.cutterTestAreas, testArea)
				end
				i = i + 1
			end
			entry.allowsForageGrowhtState = Utils.getNoNil(getXMLBool(self.xmlFile, key.."#allowsForageGrowhtState"), false)
			entry.cutterMovingDirection = Utils.sign(Utils.getNoNil(getXMLInt(self.xmlFile, key.."#movingDirection"), 1))
			table.insert(self.cutters, entry)
			j = j + 1
		end
		self.currentInputFruitType = FruitUtil.FRUITTYPE_UNKNOWN
	end
	self.setEffectEnabled = reelEffects.setEffectEnabled
end

function reelEffects:delete()
	if self.isClient then
		for i = 1, #self.cutters do
			ParticleUtil.deleteParticleSystems(self.cutters[i].threshingParticleSystems)
			for _, effectAttribute in pairs(self.cutters[i].cutterEffects) do
				EffectManager:deleteEffects(effectAttribute.effect)
			end
		end
    end
end

function reelEffects:readStream(streamId, connection)
	for i = 1, #self.cutters do
		for _, testArea in ipairs(self.cutters[i].cutterTestAreas) do
			testArea.hasFruitContact = streamReadBool(streamId)
		end
	end
end

function reelEffects:writeStream(streamId, connection)
	for i = 1, #self.cutters do
		for _, testArea in ipairs(self.cutters[i].cutterTestAreas) do
			streamWriteBool(streamId, testArea.hasFruitContact)
		end
	end
end

function reelEffects:readUpdateStream(streamId, timestamp, connection)
	if connection:getIsServer() then
		for i = 1, #self.cutters do
			for k, testArea in ipairs(self.cutters[i].cutterTestAreas) do
				testArea.hasFruitContact = streamReadBool(streamId)
			end
		end
    end
end

function reelEffects:writeUpdateStream(streamId, connection, dirtyMask)
	if not connection:getIsServer() then
		for i = 1, #self.cutters do
			for k, testArea in ipairs(self.cutters[i].cutterTestAreas) do
				streamWriteBool(streamId, testArea.hasFruitContact)
			end
		end
    end
end

function reelEffects:updateTick(dt)
	for i = 1, #self.cutters do
		if self:getIsTurnedOn() and self.movingDirection == self.cutters[i].cutterMovingDirection and not self.showFieldNotOwnedWarning then
			local hasGroundContact, _ = self:getIsTypedWorkAreaActive(WorkArea.AREATYPE_MOWER)
			if hasGroundContact then
				if self.isClient then
					local minEffectValue = math.huge
					local maxEffectValue = -math.huge
					for k, testArea in pairs(self.cutters[i].cutterTestAreas) do
						local x,y,z = getWorldTranslation(testArea.start)
						local x1,y1,z1 = getWorldTranslation(testArea.width)
						local x2,_,z2 = getWorldTranslation(testArea.height)
						if self.isServer and self:getLastSpeed() > 0.5 and self.currentInputFruitType ~= nil then
							if self:getIsTestAreaActive(testArea) then
								local fruitValue, _, growthState = Utils.getFruitArea(self.currentInputFruitType, x, z, x1, z1, x2, z2, nil, self.cutters[i].allowsForageGrowhtState)
								testArea.hasFruitContact = fruitValue > 0
								if fruitValue > 0 then
									self.cutters[i].currentInputFruitTypeGrowthState = growthState
								end
							end
						end
						if testArea.hasFruitContact then
							local lx1,_,_ = worldToLocal(self.cutters[i].cutterEffectBaseNode, x,y,z)
							local lx2,_,_ = worldToLocal(self.cutters[i].cutterEffectBaseNode, x1,y1,z1)
							minEffectValue = math.min(minEffectValue, lx1, lx2)
							maxEffectValue = math.max(maxEffectValue, lx1, lx2)
						end
					end
					local reset = false
					if minEffectValue == math.huge and maxEffectValue == -math.huge then
						minEffectValue = 0
						maxEffectValue = 0
						reset = true
					end
					if self.cutters[i].cutterMovingDirection > 0 then
						minEffectValue = minEffectValue * -1
						maxEffectValue = maxEffectValue * -1
						if maxEffectValue < minEffectValue then
							local t = minEffectValue
							minEffectValue = maxEffectValue
							maxEffectValue = t
						end
					end
					if self.currentInputFruitType ~= nil and self:getLastSpeed() > 0.5 then
						local newEffect = self.cutters[i].cutterEffects[self.currentInputFruitType]
						if newEffect ~= nil then
							if self.cutters[i].currentCutterEffect ~= newEffect then
								EffectManager:resetEffects(self.cutters[i].currentCutterEffect)
							end
							self.cutters[i].currentCutterEffect = newEffect
							for _, effect in pairs(self.cutters[i].currentCutterEffect) do
								if effect.setGrowthState ~= nil then
									effect:setGrowthState(self.cutters[i].currentInputFruitTypeGrowthState)
								end
								if effect.setMinMax ~= nil then
									effect:setMinMax(minEffectValue, maxEffectValue, reset)
								end
							end
						else
							if self.cutters[i].currentCutterEffect ~= nil then
								EffectManager:resetEffects(self.cutters[i].currentCutterEffect)
							end
						end
						for _, ps in pairs(self.cutters[i].threshingParticleSystems) do
							ParticleUtil.setEmittingState(ps, self.currentInputFruitType ~= FruitUtil.FRUITTYPE_UNKNOWN)
						end
					else
						if self.cutters[i].currentCutterEffect ~= nil then
							for _, effect in pairs(self.cutters[i].currentCutterEffect) do
								if effect.setMinMax ~= nil then
									effect:setMinMax(0, 0, true)
								end
							end
						end
						for _, ps in pairs(self.cutters[i].threshingParticleSystems) do
							ParticleUtil.setEmittingState(ps, false)
						end
						self.currentInputFruitType = FruitUtil.FRUITTYPE_UNKNOWN
					end
				end
			end
		else
			if self.cutters[i].currentCutterEffect ~= nil then
				for _, effect in pairs(self.cutters[i].currentCutterEffect) do
					if effect.setMinMax ~= nil then
						effect:setMinMax(0, 0, true)
					end
				end
				for _, ps in pairs(self.cutters[i].threshingParticleSystems) do
					ParticleUtil.setEmittingState(ps, false)
				end
				self.currentInputFruitType = FruitUtil.FRUITTYPE_UNKNOWN
			end
		end
		if self.cutters[i].cutterTurnedOnScrollers then
			Utils.updateScrollers(self.cutters[i].cutterTurnedOnScrollers, dt, self:getIsActive() and self:getIsTurnedOn())
		end
	end
end

function reelEffects:draw()
end

function reelEffects:update(dt)
end

function reelEffects:mouseEvent(posX, posY, isDown, isUp, button)
end

function reelEffects:keyEvent(unicode, sym, modifier, isDown)
end

function reelEffects:onTurnedOn()
	if self.isClient then
		for i = 1, #self.cutters do
			if self.cutters[i].knifeAnimation then
				self:playAnimation(self.cutters[i].knifeAnimation, self.cutters[i].knifeAnimationSpeed, nil, true)
			end
		end
	end
end

function reelEffects:onTurnedOff()
    if self.isClient then
		for i = 1, #self.cutters do
			if self.cutters[i].currentCutterEffect ~= nil then
				EffectManager:resetEffects(self.cutters[i].currentCutterEffect)
			end
			for _, ps in pairs(self.cutters[i].threshingParticleSystems) do
				ParticleUtil.setEmittingState(ps, false)
			end
			if self.cutters[i].knifeAnimation then
				self:stopAnimation(self.cutters[i].knifeAnimation, true)
			end
		end
		self.currentInputFruitType = FruitUtil.FRUITTYPE_UNKNOWN
	end
end

function reelEffects:loadTestAreaFromXML(testArea, xmlFile, key)
    local start = Utils.indexToObject(self.components, getXMLString(xmlFile, key .. "#startIndex"))
    local width = Utils.indexToObject(self.components, getXMLString(xmlFile, key .. "#widthIndex"))
    local height = Utils.indexToObject(self.components, getXMLString(xmlFile, key .. "#heightIndex"))
    if start ~= nil and width ~= nil and height ~= nil then
        testArea.start = start
        testArea.width = width
        testArea.height = height
        testArea.hasFruitContact = false
        return true
    end
    return false
end

function reelEffects:getIsTestAreaActive(area)
    return true
end

function reelEffects:setEffectEnabled(mowerEffect, isActive)

    mowerEffect.isActive = isActive

    if self.isClient then

        if isActive then

            EffectManager:setFillType(mowerEffect.effects, self.currentWindrowType or FillUtil.FILLTYPE_GRASS_WINDROW)

            EffectManager:startEffects(mowerEffect.effects)

        else

            EffectManager:stopEffects(mowerEffect.effects)

        end

    end

end